package com.zanxiang.game.data.serve.service.impl;

import com.zanxiang.erp.security.util.SecurityUtil;
import com.zanxiang.game.data.serve.pojo.dto.OverallSummaryDTO;
import com.zanxiang.game.data.serve.pojo.entity.AdsGamePushNum;
import com.zanxiang.game.data.serve.pojo.entity.AdsOverallSummary;
import com.zanxiang.game.data.serve.pojo.vo.OverallSummaryLineDataVO;
import com.zanxiang.game.data.serve.pojo.vo.OverallSummaryVO;
import com.zanxiang.game.data.serve.service.IOverallSummaryService;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;
import org.nutz.dao.Cnd;
import org.nutz.dao.Condition;
import org.nutz.dao.Dao;
import org.nutz.dao.Sqls;
import org.nutz.dao.sql.Criteria;
import org.nutz.dao.sql.Sql;
import org.springframework.beans.BeanUtils;
import org.springframework.beans.BeanWrapper;
import org.springframework.beans.BeanWrapperImpl;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import java.beans.PropertyDescriptor;
import java.math.BigDecimal;
import java.math.RoundingMode;
import java.time.LocalDate;
import java.time.temporal.ChronoUnit;
import java.util.*;

/**
 * @author tianhua
 * @time 2023/7/4
 * @Description 整体概况逻辑处理
 **/
@Service
@Slf4j
public class OverallSummaryServiceImpl implements IOverallSummaryService {

    @Autowired
    private Dao dao;

    @Override
    public OverallSummaryVO getOverallSummaryTotalData(OverallSummaryDTO dto) {

        //获取查询条件 整体数据查询 不需要日期 需要gameId
        Condition totalDataCon = myCondition(dto, true,false, -1);
        //查询条件转换为sql语句 设置映射 Entity为 AdsOverallSummary对象
        String totalDataConString = totalDataCon.toSql(dao.getEntity(AdsOverallSummary.class));
        //自定义sql语句
        Sql totalDataSql = Sqls.create("""
                SELECT
                IFNULL(SUM(x.today_player_count),0) total_players ,
                IFNULL(SUM(x.today_total_cost),0) total_cost ,
                IFNULL(SUM(x.today_amount),0) total_amount ,
                IFNULL(SUM(x.today_agent_count),0) total_agent_count ,
                IFNULL(SUM(x.today_amount)/SUM(x.today_total_cost),0) total_roi
                FROM ads_overall_summary x
                """ + totalDataConString);

        //自定义回显对象 结果自动封装到给定的Entity对象中
        totalDataSql.setCallback(Sqls.callback.entity());
        totalDataSql.setEntity(dao.getEntity(OverallSummaryVO.class));
        //执行自定义sql语句
        dao.execute(totalDataSql);
        //得到一个包含有总数据的OverallSummaryVO对象
        OverallSummaryVO totalDataVO = totalDataSql.getObject(OverallSummaryVO.class);

        //获取查询条件 查询今日数据 需要按当前日期查询
        Condition todayDataCon = myCondition(dto,true, false, 1);
        //查询条件转换为sql语句 设置映射 Entity为 AdsOverallSummary对象
        String todayDataConString = todayDataCon.toSql(dao.getEntity(AdsOverallSummary.class));
        Sql todayDataSql = Sqls.create("""
                SELECT
                IFNULL(SUM(x.today_player_count),0) today_player_count ,
                IFNULL(SUM(x.yesterday_player_count),0) yesterday_player_count ,
                IFNULL(SUM(x.today_total_cost),0) today_total_cost ,
                IFNULL(SUM(x.yesterday_total_cost),0) yesterday_total_cost ,
                IFNULL(SUM(x.today_new_player_amount),0) today_new_player_amount ,
                IFNULL(SUM(x.today_amount),0) today_amount ,
                IFNULL(SUM(x.yesterday_new_player_amount),0) yesterday_new_player_amount ,
                IFNULL(SUM(x.yesterday_amount),0) yesterday_amount ,
                IFNULL(SUM(x.today_agent_count),0) today_agent_count ,
                IFNULL(SUM(x.yesterday_agent_count),0) yesterday_agent_count ,
                IFNULL(SUM(x.first_new_user_amount)/SUM(x.today_total_cost),0) first_roi
                FROM ads_overall_summary x
                """ + todayDataConString);

        //自定义回显对象 结果自动封装到给定的Entity对象中
        todayDataSql.setCallback(Sqls.callback.entity());
        todayDataSql.setEntity(dao.getEntity(OverallSummaryVO.class));
        //执行自定义sql语句
        dao.execute(todayDataSql);
        //得到一个包含有今日和昨日数据的OverallSummaryVO对象
        OverallSummaryVO overallSummaryVO = todayDataSql.getObject(OverallSummaryVO.class);

        //拼接参数
        copyNullProperties(totalDataVO, overallSummaryVO);

        //7日和30日Roi数据需要手动计算
        //7日内新用户累计充值
        BigDecimal d7NewPlayerAmount = (BigDecimal) dao.func2(AdsOverallSummary.class, "sum", "todayAmount", myCondition(dto,true, false, 7));
        //7日内总消耗
        BigDecimal d7TotalCost = (BigDecimal) dao.func2(AdsOverallSummary.class, "sum", "todayTotalCost", myCondition(dto, true,false, 7));
        //cost 查询出的结果为null或者为0.00 设置为默认值0
        overallSummaryVO.setD7TotalRoi((d7TotalCost == null || BigDecimal.ZERO.compareTo(d7TotalCost) == 0) ? BigDecimal.ZERO : d7NewPlayerAmount.multiply(BigDecimal.valueOf(100)).divide(d7TotalCost, 4, RoundingMode.HALF_UP));

        //30日内新用户累计充值
        BigDecimal d30NewPlayerAmount = (BigDecimal) dao.func2(AdsOverallSummary.class, "sum", "todayAmount", myCondition(dto,true, false, 30));
        //30日内总消耗
        BigDecimal d30TotalCost = (BigDecimal) dao.func2(AdsOverallSummary.class, "sum", "todayTotalCost", myCondition(dto, true,false, 30));
        //计算30日Roi
        //cost 查询出的结果为null或者为0.00 设置为默认值0
        overallSummaryVO.setD30TotalRoi((d30TotalCost == null || BigDecimal.ZERO.compareTo(d30TotalCost) == 0) ? BigDecimal.ZERO : d30NewPlayerAmount.multiply(BigDecimal.valueOf(100)).divide(d30TotalCost, 4, RoundingMode.HALF_UP));

        //计算百分比
        overallSummaryVO.setFirstRoi(overallSummaryVO.getFirstRoi().multiply(BigDecimal.valueOf(100)));
        overallSummaryVO.setTotalRoi(overallSummaryVO.getTotalRoi().multiply(BigDecimal.valueOf(100)));

        //计算游戏数量相关数据
        //查询总投游戏数量
        Sql gameCountTotalSql = Sqls.create(gameCountSql() +
                myCondition(dto,false,false,-1).toSql(dao.getEntity(AdsGamePushNum.class))
        );
        //自定义回显对象 结果自动封装到给定的Entity对象中
        gameCountTotalSql.setCallback(Sqls.callback.integer());
        //执行自定义sql语句
        dao.execute(gameCountTotalSql);
        overallSummaryVO.setTotalGameCount(Long.valueOf(gameCountTotalSql.getInt()));

        //查询今日在投游戏数量
        Sql gameCountTodaySql = Sqls.create(gameCountSql() +
                myCondition(dto,false,false,1).toSql(dao.getEntity(AdsGamePushNum.class))
        );
        //自定义回显对象 结果自动封装到给定的Entity对象中
        gameCountTodaySql.setCallback(Sqls.callback.integer());
        //执行自定义sql语句
        dao.execute(gameCountTodaySql);
        overallSummaryVO.setTodayGameCount(Long.valueOf(gameCountTodaySql.getInt()));

        //查询昨日在投游戏数量
        Sql gameCountYesterdaySql = Sqls.create(gameCountSql() +
                myCondition(dto,false,false,0).toSql(dao.getEntity(AdsGamePushNum.class))
        );
        //自定义回显对象 结果自动封装到给定的Entity对象中
        gameCountYesterdaySql.setCallback(Sqls.callback.integer());
        //执行自定义sql语句
        dao.execute(gameCountYesterdaySql);
        overallSummaryVO.setYesterdayGameCount(Long.valueOf(gameCountYesterdaySql.getInt()));

        return overallSummaryVO;
    }

    @Override
    public List<OverallSummaryLineDataVO> getOverallSummaryLineData(OverallSummaryDTO dto) {

        //根据条件获取数据
        Condition lineCon = myCondition(dto,true, true, -1);
        //查询条件变为字符串
        String lineConString = lineCon.toSql(dao.getEntity(AdsOverallSummary.class));
        //自定义sql语句 加上条件字符串
        Sql sql = Sqls.create("""
                Select
                dt,
                SUM(x.today_total_cost) cost,
                SUM(x.today_amount) amount,
                SUM(x.today_buy_amount) buy_amount,
                SUM(x.today_nature_amount) nature_amount,
                IFNULL(SUM(x.first_new_user_amount)/SUM(x.today_total_cost),0) first_roi,
                SUM(x.today_player_count) player_account
                from ads_overall_summary x
                """ + lineConString);

        //自定义回显对象 结果自动添加到一个List中 List<Entity>
        sql.setCallback(Sqls.callback.entities());
        sql.setEntity(dao.getEntity(OverallSummaryLineDataVO.class));

        //执行自定义sql语句
        dao.execute(sql);

        //判断查询日期的数据是否都存在 结果返回
        return checkAllDateData(sql.getList(OverallSummaryLineDataVO.class), dto);
    }

    /**
     * 自定义查询条件
     *
     * @param dto         前端传递的查询参数实体
     * @param needGameId 是否需要拼接gameId查询条件
     * @param needLocalDate 是否拼接日期查询条件
     * @param queryDays     拼接需要查询数据的天数条件（给今日昨日、首日Roi、7日Roi以及30日Roi使用）
     * @return Condition 查询条件
     */
    private Condition myCondition(OverallSummaryDTO dto,Boolean needGameId, Boolean needLocalDate, Integer queryDays) {

        Criteria cri = Cnd.cri();
        //默认查询条件
        if (dto.getGameId() != null && needGameId) {
            //拼接游戏查询条件
            cri.where().andEquals("gameId", dto.getGameId());
        }
        if (dto.getAgentId() != null) {
            //拼接渠道查询条件
            cri.where().andEquals("agentId", dto.getAgentId());
        }
        if (dto.getAccountId() != null) {
            //拼接广告账户查询条件
            cri.where().andEquals("accountId", dto.getAccountId());
        }
        if (StringUtils.isNotBlank(dto.getSourceSystem())){
            //拼接查询 sdk 来源条件
            cri.where().andEquals("sourceSystem", dto.getSourceSystem());
        }
        //前端传递 pitcherId 根据 pitcherId 查询
        if (dto.getPitcherId() != null) {
            //拼接投手查询条件
            cri.where().andEquals("pitcherId", dto.getPitcherId());
        } else if (SecurityUtil.isManager()) {
            //前端没有传递 pitcherId 参数并且是管理员 查看所有数据 排除掉 pitcherId = -2 的数据
            cri.where().andNotEquals("pitcherId", -2);
        } else {
            //前端没有传递 pitcherId 并且不是管理员 只能看自己的数据
            cri.where().andEquals("pitcherId", SecurityUtil.getUserId());
        }

        //根据参数拼接条件
        switch (queryDays) {
            //前一天时间
            case 0 -> cri.where().andEquals("dt", LocalDate.now().minusDays(1));
            //当天时间
            case 1 -> cri.where().andEquals("dt", LocalDate.now());
            //7天时间
            case 7 -> cri.where().andBetween("dt", LocalDate.now().minusDays(7), LocalDate.now());
            //30天时间
            case 30 -> cri.where().andBetween("dt", LocalDate.now().minusDays(30), LocalDate.now());
            default -> {
            }
        }

        //折线数据需要拼接时间查询条件
        if (needLocalDate) {
            //如果传入时间区间参数为空 设置默认值
            if (dto.getBeginDate() == null || dto.getEndDate() == null) {
                dto.setBeginDate(LocalDate.now().minusDays(30));
                dto.setEndDate(LocalDate.now());
            }
            //添加日期条件
            cri.where().andBetween("dt", dto.getBeginDate(), dto.getEndDate());
            //按日期分组
            cri.getGroupBy().groupBy("dt");
            //按日期排序 升序
            cri.getOrderBy().asc("dt");
        }

        return cri;
    }

    /**
     * 所有为空值的属性都不copy
     *
     * @param source 原数据
     * @param target 目标数据
     */
    private void copyNullProperties(Object source, Object target) {
        BeanUtils.copyProperties(source, target, getNullField(source));
    }

    /**
     * 获取属性中为空的字段
     *
     * @param target 目标对象
     * @return 不需要替换的字段数组
     */
    private static String[] getNullField(Object target) {
        BeanWrapper beanWrapper = new BeanWrapperImpl(target);
        PropertyDescriptor[] propertyDescriptors = beanWrapper.getPropertyDescriptors();
        Set<String> notNullFieldSet = new HashSet<>();
        if (propertyDescriptors.length > 0) {
            for (PropertyDescriptor p : propertyDescriptors) {
                String name = p.getName();
                Object value = beanWrapper.getPropertyValue(name);
                if (Objects.isNull(value)) {
                    notNullFieldSet.add(name);
                }
            }
        }
        String[] notNullField = new String[notNullFieldSet.size()];

        return notNullFieldSet.toArray(notNullField);
    }

    /**
     * 检查前端查询日期内的数据是否在数据库中都存在
     * 不存在要设置默认值 0
     *
     * @param list  需要判断的list
     * @param dto 前端传递的查询参数实体
     * @return 返回给前端的折线图数据 list
     */
    private List<OverallSummaryLineDataVO> checkAllDateData(List<OverallSummaryLineDataVO> list, OverallSummaryDTO dto) {

        //先判断是否缺少数据 不少数据直接返回
        if ((list.size() - 1) == dto.getBeginDate().until(dto.getEndDate(), ChronoUnit.DAYS)) {
            return list;
        }

        //创建一个Map用来存储日期数据
        Map<LocalDate, Integer> map = new HashMap<>();
        //添加日期数据
        LocalDate tempDate = dto.getBeginDate();
        while (!tempDate.isAfter(dto.getEndDate())) {
            map.put(tempDate, 1);
            tempDate = tempDate.plusDays(1);
        }
        //循环遍历list 查询vo里的日期参数
        for (OverallSummaryLineDataVO vo : list) {
            Integer value = map.get(vo.getDt());
            if (value != null) {
                map.put(vo.getDt(), ++value);
            }
        }
        //遍历map数据 value为 1 的表示 list中没有该日期的实体
        for (Map.Entry<LocalDate, Integer> entry : map.entrySet()) {
            if (entry.getValue() == 1) {
                //list数据添加一个默认VO对象
                list.add(OverallSummaryLineDataVO.builder()
                        .dt(entry.getKey())
                        .amount(BigDecimal.ZERO)
                        .buyAmount(BigDecimal.ZERO)
                        .natureAmount(BigDecimal.ZERO)
                        .cost(BigDecimal.ZERO)
                        .firstRoi(BigDecimal.ZERO)
                        .playerAccount(0L)
                        .build());
            }
        }
        //排序 lambda表达式 按日期升序排序
        list.sort((vo1, vo2) -> vo1.getDt().isBefore(vo2.getDt()) ? -1 : 1);

        return list;
    }

    /**
     * 共用的gameCount sql语句
     * @return String
     */
    private String gameCountSql(){
        return """
               SELECT
                   COUNT(DISTINCT a.game_id)
               FROM
                   game_ads.ads_game_push_num a
               """;
    }

}
